"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const awayback_model_1 = require("./awayback.model");
const lodash_1 = require("lodash");
/**
 * @license
 * awayback
 * Released under MIT license
 * Copyright Luca Raúl Joos
 */
function awayback() {
    const events = {};
    function create(event) {
        events[event] = {
            callbacks: new Proxy([], {
                set: (target, property, value, receiver) => {
                    const self = events[event];
                    if (!self)
                        return false;
                    if (self.runs > 0) {
                        self.callbacks.forEach((callback) => {
                            let isExiting = false;
                            while (callback.runs < self.runs && !isExiting && callback.options.isExecutingPrevious) {
                                if (callback.type === awayback_model_1.CallbackType.on ||
                                    (callback.type === awayback_model_1.CallbackType.once && callback.runs === 0) ||
                                    (callback.type === awayback_model_1.CallbackType.only &&
                                        callback.runs === 0 &&
                                        self.callbacks.reduce((sum, callback) => sum + callback.runs, 0) === 0)) {
                                    callback.runs++;
                                    callback.handler(...self.data[callback.runs - 1]);
                                }
                                else {
                                    isExiting = true;
                                }
                            }
                        });
                    }
                    return Reflect.set(target, property, value, receiver);
                },
            }),
            runs: 0,
            data: [],
        };
    }
    function emit(event, ...data) {
        if (typeof events[event] === 'undefined')
            create(event);
        const self = events[event];
        if (!self)
            return;
        if (typeof data !== 'undefined') {
            self.data.push(data);
        }
        self.runs++;
        self.callbacks.forEach((callback) => {
            if (callback.type === awayback_model_1.CallbackType.on ||
                (callback.type === awayback_model_1.CallbackType.once && callback.runs === 0) ||
                (callback.type === awayback_model_1.CallbackType.only &&
                    callback.runs === 0 &&
                    self.callbacks.reduce((sum, callback) => sum + callback.runs, 0) === 0)) {
                callback.runs++;
                callback.handler(...self.data[callback.runs - 1]);
            }
        });
    }
    function listen(type, event, handler, options) {
        if (typeof events[event] === 'undefined')
            create(event);
        const self = events[event];
        if (!self)
            return;
        self.callbacks.push({
            handler: handler,
            type,
            runs: 0,
            options: (0, lodash_1.merge)({ isExecutingPrevious: false }, options !== null && options !== void 0 ? options : {}),
        });
    }
    function on(event, handler, options) {
        listen(awayback_model_1.CallbackType.on, event, handler, options);
    }
    function once(event, handler, options) {
        listen(awayback_model_1.CallbackType.once, event, handler, options);
    }
    function only(event, handler, options) {
        listen(awayback_model_1.CallbackType.only, event, handler, options);
    }
    function remove(event, handler) {
        if (typeof events[event] === 'undefined')
            return;
        const self = events[event];
        if (!self)
            return;
        self.callbacks = self.callbacks.filter((callback) => callback.handler !== handler);
    }
    return {
        events,
        emit,
        on,
        once,
        only,
        remove,
    };
}
exports.default = awayback;
